;; -*- Scheme -*-
;;
;;	Author: Sam Rushing <rushing@nightmare.com>
;;	$Id: select-coroutine.scm,v 1.4 1997/03/09 20:59:24 rushing Exp rushing $
;;

(include "eotya.scm")

(define-operation (wait self socket event))
(define-operation (spin self))
(define-operation (spawn self thunk))

;; a <selector> provides the context for a select() loop.  it contains
;; three ready-made <fd_set> objects and a simple interface for selecting
;; on objects (as opposed to file descriptors).

;; <wheel> adds a 'thread-like' capability, using coroutines
;; implemented with call/cc.  socket objects that are 'chained' to it
;; will suspend execution whenever a socket operation would block;
;; storing their continuation in a dictionary for later resumption.
;; The select loop is kicked off by calling the 'spin' method.

(define (make-wheel selector-size selector-timeout)
  (let ((wheel-continuation-stack (make-stack))
	(read-registry	(make-dictionary))
	(write-registry	(make-dictionary))
	(expt-registry	(make-dictionary)))

    (define (get-registry tag)
      (case tag
	((read)	read-registry)
	((write) write-registry)
	((expt)	expt-registry)))

    (object-with-ancestors
     ((selector (make-selector selector-size)))

     ;; --------------------------------------------------
     ;; spawn
     ;; --------------------------------------------------

     ((spawn self thunk)
      (call-with-current-continuation
       (lambda (continuation)
	 (push! wheel-continuation-stack continuation)
	 (thunk)
	 ((pop! wheel-continuation-stack) #f))))

     ;; --------------------------------------------------
     ;; wait
     ;; --------------------------------------------------

     ((wait self socket event)
      (call-with-current-continuation
       (lambda (chain-continuation)
	 (set-item! (get-registry event) socket chain-continuation)
	 ((pop! wheel-continuation-stack) #f)))
      ;; when we get back, first thing we do is remove
      ;; this socket from the event registry
      (del-item! (get-registry event) socket))

     ;; --------------------------------------------------
     ;; spin
     ;; --------------------------------------------------
     ;; the main loop

     ((spin self)

      (let loop ()
	(cond ((> (apply + (map size (list read-registry write-registry expt-registry))) 0)
	       (let ((fired (operate-as
			     selector select self
			     (keys read-registry)
			     (keys write-registry)
			     (keys expt-registry)
			     selector-timeout)))
		 ;; <fired> is a triple of (read-objects, write-objects, expt-objects)
		 (let ((read-continuations
			(map
			 (lambda (x) (cdr (get-item read-registry x)))
			 (list-ref fired 0)))
		       (write-continuations
			(map
			 (lambda (x) (cdr (get-item write-registry x)))
			 (list-ref fired 1)))
		       (expt-continuations
			(map
			 (lambda (x) (cdr (get-item expt-registry x)))
			 (list-ref fired 2))))
		   ;; flatten the three lists of continuations
		   (let ((continuations
			  (apply append
				 (list read-continuations
				       write-continuations
				       expt-continuations))))
		     (for-each
		      (lambda (chain-continuation)
			(call-with-current-continuation
			 (lambda (c)
			   (push! wheel-continuation-stack c)
			   (chain-continuation #f))))
		      continuations))))
	       (loop))
	      ))))))

;; A 'chained' socket is a socket object specifically meant for use
;; with a 'wheel'.  It is initialized to non-blocking mode.  Whenever
;; a socket operation blocks (usually by returning EWOULDBLOCK), the
;; 'chain' is suspended.  The wheel will resume it when the approriate
;; event is detected by a call to select().

(define (make-chained-socket wheel type family protocol . provided-socket)
  (object-with-ancestors
   ;; parent
   ((socket
     (let ((s (cond ((null? provided-socket)
		     (make-socket type family protocol))
		    (else
		     (car provided-socket)))))
       ;; make sure the socket is in non-blocking mode
       (set-blocking s #f)
       s)))

   ;; methods:
   ;; override any blocking operations

   ((connect self address)
    (let ((result (operate-as socket connect self address)))
      (if (not (= result -1))
	  result
	  (wait wheel self 'write))))

   ((accept self)
    (let loop ((client (operate-as socket accept self)))
      (if client
	  (make-chained-socket wheel type family protocol client)
	  (begin
	    (wait wheel self 'read)
	    (loop (operate-as socket accept self))))))

   ((recv self buffer-size)
    (let loop ((result (operate-as socket recv self buffer-size)))
      (if result
	  result
	  (begin
	    (wait wheel self 'read)
	    (loop (operate-as socket recv self buffer-size))))))

   ((send self buffer)
    (let loop ((result (operate-as socket send self buffer)))
      (if (> result 0)
	  result
	  (begin
	    (wait wheel self 'write)
	    (loop (operate-as socket send self buffer))))))
   ))
